package dfki.rdf.utils;

import java.util.*;
import java.io.*;
import java.lang.reflect.*;

import org.w3c.rdf.model.*;
import org.w3c.rdf.syntax.*;
import org.w3c.rdf.util.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.*;
import org.w3c.rdf.vocabulary.rdf_schema_19990303.*;


public class RDFExport {

  RDFFactory _rdfFactory;
  NodeFactory _nodeFactory;
  Model _model;

  String _anonNamespace = "http://www.anonymous.org/rdf#";
  String _sysNamespace;

  final static String _damlNamespace = "http://www.daml.org/2001/03/daml+oil#";
  Resource _damlList;
  Resource _damlFirst;
  Resource _damlRest;
  Resource _damlNil;

  HashMap _exportedObjects; // maps objects to (resource) names
  int _genid;


  public RDFExport(String namespace) {
    _sysNamespace = namespace;
  }


  // RDF model/statement/resource handling

  void add(Statement statement) {
    if (statement != null) {
      try {
        _model.add(statement);
      } catch (Exception e) { System.err.println(e); }
    }
  }

  Statement statement(Resource subject, Resource predicate, 
      RDFNode object) {
    if (subject != null && predicate != null && object != null) {
      try {
        return _nodeFactory.createStatement(subject, predicate, object);
      } catch (Exception e) { System.err.println(e); }
    }
    return null;
  }

  Resource resource(String namespace, String name) {
    try {
      return _nodeFactory.createResource(namespace, name);
    } catch (Exception e) { System.err.println(e); }
    return null;
  }

  Resource anonResource(String name) {
    return resource(_anonNamespace, name);
  }

  Resource sysResource(String name) {
    return resource(_sysNamespace, name);
  }

  Resource liResource(int i) {
    return resource(RDF._Namespace, "_" + i);
  }

  Literal literal(String string) {
    try {
      return _nodeFactory.createLiteral(string);
    } catch (Exception e) { System.err.println(e); }
    return null;
  }

  String getGenid() {
    return "genid" + _genid++;
  }


  // export

  public void exportObjects(Collection objects, String filename) {
    // init
    _genid = 1;
    _rdfFactory = new RDFFactoryImpl();
    _model = _rdfFactory.createModel();
    try {
      _nodeFactory = _model.getNodeFactory();
    } catch (Exception e) {
      System.err.println(e);
      return;
    }
    // export
    _exportedObjects = new HashMap();
    for (Iterator objectIterator = objects.iterator();
	 objectIterator.hasNext();) {
      Object object = objectIterator.next();
      exportObject(object);
    }
    // save model
    try {
      RDFUtil.saveModel(_model, filename, 
	new org.w3c.rdf.implementation.syntax.sirpac.SiRS());
    } catch (Exception e) {
      System.err.println(e);
    }
  }

  public String exportObject(Object object) {
    String name = (String)_exportedObjects.get(object);
    if (name == null) { // not yet exported
      name = getGenid();
      _exportedObjects.put(object, name);
      Class cls = object.getClass();
      exportType(name, cls);
      Method[] methods = cls.getMethods();
      exportPropertyValues(name, object, methods);
    }
    return name;
  }

  void exportType(String name, Class cls) {
    String className = getClassName(cls);
    add(statement(anonResource(name), RDF.type, sysResource(className)));
  }

  String getClassName(Class cls) {
    String className = cls.getName();
    int p = className.lastIndexOf('.');
    return className.substring(p+1);
  }

  void exportPropertyValues(String name, Object object, Method[] methods) {
    for (int i = 0; i < methods.length; i++) {
      Method method = methods[i];
      String methodName = method.getName();
      if (methodName.startsWith("get") && !methodName.equals("getClass")) {
	Class[] parameterTypes = method.getParameterTypes();
	if (parameterTypes.length == 0) {
	  Object value = null;
	  try {
	    value = method.invoke(object, null);
	  } catch (Exception e) {
	    System.err.println("exportPropertyValues: "  + e);
	  }
	  String propertyName = getPropertyName(methodName);
	  exportPropertyValue(name, propertyName, value);
	}
      }
    }
  }

  void exportPropertyValue(String name, String propertyName, Object value) {
    if (value != null) {
      if (value instanceof Collection) {
	if (value instanceof List) { // list -> Seq (or daml:List)
	  exportValueList(name, propertyName, (List)value);
	} else { // set -> multiple occurences of proptery
	  exportValueSet(name, propertyName, (Collection)value);
	}
      } else { // simple value
	exportSimpleValue(name, propertyName, value);
      }
    }
  }

  void exportValueList(String name, String propertyName, List values) {
    String seqName = getGenid();
    Resource seqResource = anonResource(seqName);
    add(statement(anonResource(name), sysResource(propertyName), seqResource));
    add(statement(seqResource, RDF.type, RDF.Seq));
    int i = 1;
    for (Iterator valueIterator = values.iterator();
	 valueIterator.hasNext();) {
      Object value = valueIterator.next();
      exportSimpleValue(seqResource, liResource(i), value);
      i++;
    }
  }

  void exportValueSet(String name, String propertyName, Collection values) {
    for (Iterator valueIterator = values.iterator();
	 valueIterator.hasNext();) {
      Object value = valueIterator.next();
      exportSimpleValue(name, propertyName, value);
    }
  }

  void exportSimpleValue(String name, String propertyName, Object value) {
    exportSimpleValue(anonResource(name), sysResource(propertyName), value);
  }

  void exportSimpleValue(Resource subject, Resource property, Object value) {
    if (value instanceof String) { // rdf:Literal
      add(statement(subject, property, literal((String)value)));
    } else { // resource
      String valueName = exportObject(value);
      add(statement(subject, property, anonResource(valueName)));
    }
  }

  String getPropertyName(String methodName) {
    // getProp -> prop
    // get_prop -> prop and esp. get_Prop -> Prop (upper case properties)
    if (methodName.startsWith("get_"))
      return methodName.substring(4);
    else
      return Character.toLowerCase(methodName.charAt(3)) 
	+ methodName.substring(4);
  }


}


